library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity Audio_PWM is
  port (
    clk, rst: in std_logic;
    D_music: in std_logic_vector(15 downto 0); -- Data word = 8-bits left + 8-bits right channel
    D_effects: in std_logic_vector(15 downto 0); -- Data word = 8-bits left + 8-bits right channel
    Left_PWM_Q: out std_logic;
    Right_PWM_Q: out std_logic;
	FIFO_music_wr: in std_logic;
	FIFO_music_full: out std_logic;
	FIFO_effects_wr: in std_logic;
	FIFO_effects_full: out std_logic
  );
end;

architecture distributed_pulses of Audio_PWM is
 
  signal D_music_buffered: std_logic_vector(15 downto 0);
  signal D_effects_buffered: std_logic_vector(15 downto 0);

  signal acc_left_audio: unsigned (9 downto 0);  -- one bit bigger than D
  signal acc_left_temp: unsigned (9 downto 0);  -- one bit bigger than D  
  
  signal acc_right_audio: unsigned (9 downto 0); -- one bit bigger than D
  signal acc_right_temp: unsigned (9 downto 0); -- one bit bigger than D  
  signal FIFO_Read_clk: std_logic;
  
  signal sample_counter: unsigned (10 downto 0);
  signal positive_reset: std_logic;
  
  type audio_state is (add_op,shift_op1,shift_op2,writedata);
  signal audiostate : audio_state :=add_op;  
 


begin

-- Create FIFO buffer for the music L+R channels
music_FIFO : entity work.M68K_Audio_Musicbuf1
    port map (Data(15 downto 0)=>D_music, WrClock=> Not clk, RdClock=>FIFO_Read_clk, 
		WrEn=>FIFO_music_wr, RdEn=> '1', Reset=>positive_reset, RPReset=>positive_reset, 
		Q(15 downto 0)=>D_music_buffered, AlmostFull=> FIFO_music_full);
		
-- Create FIFO buffer for the sound-effect L+R channels
effects_FIFO : entity work.M68K_Audio_Musicbuf2
    port map (Data(15 downto 0)=>D_effects, WrClock=> Not clk, RdClock=>FIFO_Read_clk, 
		WrEn=>FIFO_effects_wr, RdEn=> '1', Reset=> positive_reset, RPReset=>positive_reset, 
		Q(15 downto 0)=>D_effects_buffered, AlmostFull=> FIFO_effects_full);

--Empty=> null,  Full=> null, AlmostEmpty=> null, AlmostFull=> null);

  positive_reset <= NOT rst;

  Left_PWM_Q <= acc_left_audio(9);   -- Left acc carry bit provides Left channel PWM output
  Right_PWM_Q <= acc_right_audio(9); -- Right acc carry bit provides Right channel PWM output
  
  FIFO_Read_clk <= sample_counter(10); -- Set for 24414.0625 KHz FIFO out rate
 
  process (clk, rst)
  begin
    if rst = '0' then
      acc_left_audio <= (others => '0');
      acc_right_audio <= (others => '0');
	  acc_left_temp <= (others => '0');
	  acc_right_temp <= (others => '0');
	  audiostate <= add_op;
	  sample_counter <= "00000000000";
	  
	elsif rising_edge(clk) then
	
		sample_counter <= sample_counter + 1;
		case audiostate is
			when add_op => 
				acc_left_temp <= unsigned('0' & D_music_buffered(7 downto 0) & '0') + 
				unsigned('0' & D_effects_buffered(7 downto 0) & '0') ;
				acc_right_temp <= unsigned('0' & D_music_buffered(15 downto 8) & '0') + 
				unsigned('0' & D_effects_buffered(15 downto 8) & '0') ;
				audiostate <= shift_op1;				
			when shift_op1 => 	
				acc_left_temp <= acc_left_temp srl 1;
				audiostate <= shift_op2;
			when shift_op2 => 	
				acc_right_temp <= acc_right_temp srl 1;
				audiostate <= writedata;
			when writedata => 			
				acc_left_audio <= ('0' & acc_left_audio(8 downto 0))  -- all but the carry
					 + acc_left_temp(8 downto 0)             -- input demand (lower 8-bits of data-in)
					 + acc_left_audio(9 downto 9);  -- the carry bit
				acc_right_audio <= ('0' & acc_right_audio(8 downto 0))  -- all but the carry
					 + acc_right_temp(8 downto 0)             -- input demand (upper 8-bits of data-in)
					 + acc_right_audio(9 downto 9);  -- the carry bit 
				audiostate <= add_op;
				
			when others =>
				null;					 
			end case;
		end if;
	end process;

end;